﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Transformations;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public class TransformationToSymbolicTreeMapper : ITransformationMapper<ISymbolicExpressionTree> {
    private ITransformation transformation;
    private string column;
    private ISymbolicExpressionTree tree;

    #region ITransformationMapper<ISymbolicExpressionTree> Members

    public ISymbolicExpressionTree GenerateModel(ITransformation transformation) {
      InitComponents(transformation);

      if (transformation is LinearTransformation) {
        return GenerateModelForLinearTransformation();
      } else if (transformation is ExponentialTransformation) {
        return GenerateModelForExponentialTransformation();
      } else if (transformation is LogarithmicTransformation) {
        return GenerateModelForLogarithmicTransformation();
      } else if (transformation is PowerTransformation) {
        return GenerateModelForPowerTransformation();
      } else if (transformation is ReciprocalTransformation) {
        return GenerateModelForReciprocalTransformation();
      } else if (transformation is ShiftStandardDistributionTransformation) {
        return GenerateModelForShiftStandardDistributionTransformation();
      }
      throw new NotImplementedException();
    }

    public ISymbolicExpressionTree GenerateInverseModel(ITransformation transformation) {
      InitComponents(transformation);

      if (transformation is LinearTransformation) {
        return GenerateInverseModelForLinearTransformation();
      } else if (transformation is ExponentialTransformation) {
        return GenerateInverseModelForExponentialTransformation();
      } else if (transformation is LogarithmicTransformation) {
        return GenerateInverseModelForLogarithmicTransformation();
      } else if (transformation is PowerTransformation) {
        return GenerateInverseModelForPowerTransformation();
      } else if (transformation is ReciprocalTransformation) {
        return GenerateInverseModelForReciprocalTransformation();
      } else if (transformation is ShiftStandardDistributionTransformation) {

      }

      throw new NotImplementedException();
    }

    #endregion

    // helper

    private ISymbolicExpressionTree GenerateModelForLinearTransformation() {
      var linearTransformation = (LinearTransformation)transformation;
      var kValue = linearTransformation.Multiplier;
      var dValue = linearTransformation.Addend;

      // k * x
      var multiplicationNode = new Multiplication().CreateTreeNode();
      var kNode = new ConstantTreeNode(new Constant() { Description = "K" }) { Value = kValue };
      var xNode = new Variable(column, "x").CreateTreeNode();
      multiplicationNode.AddSubtree(kNode);
      multiplicationNode.AddSubtree(xNode);

      // ( k * x ) + d
      var additionNode = new Addition().CreateTreeNode();
      var dNode = new ConstantTreeNode(new Constant() { Description = "d" }) { Value = dValue };
      additionNode.AddSubtree(multiplicationNode);
      additionNode.AddSubtree(dNode);

      tree.Root.AddSubtree(additionNode);
      return tree;
    }

    private ISymbolicExpressionTree GenerateInverseModelForLinearTransformation() {
      var linearTransformation = (LinearTransformation)transformation;
      var kValue = linearTransformation.Multiplier;
      var dValue = linearTransformation.Addend;

      // x - d
      var substractionNode = new Subtraction().CreateTreeNode();
      var dNode = new ConstantTreeNode(new Constant() { Description = "d" }) { Value = dValue };
      var xNode = new Variable(column, "x").CreateTreeNode();
      substractionNode.AddSubtree(xNode);
      substractionNode.AddSubtree(dNode);

      // ( x - d ) / k
      var divisionNode = new Division().CreateTreeNode();
      var kNode = new ConstantTreeNode(new Constant() { Description = "K" }) { Value = kValue };
      divisionNode.AddSubtree(substractionNode);
      divisionNode.AddSubtree(kNode);

      tree.Root.AddSubtree(divisionNode);
      return tree;
    }


    private ISymbolicExpressionTree GenerateModelForExponentialTransformation() {
      var exponentialTransformation = (ExponentialTransformation)transformation;
      var bValue = exponentialTransformation.Base;

      return GenTreePow_b_x(bValue);
    }

    private ISymbolicExpressionTree GenerateInverseModelForExponentialTransformation() {
      var exponentialTransformation = (ExponentialTransformation)transformation;
      var bValue = exponentialTransformation.Base;

      return GenTreeLog_x_b(bValue);
    }


    private ISymbolicExpressionTree GenerateModelForLogarithmicTransformation() {
      var logarithmicTransformation = (LogarithmicTransformation)transformation;
      var bValue = logarithmicTransformation.Base;

      return GenTreeLog_x_b(bValue);
    }

    private ISymbolicExpressionTree GenerateInverseModelForLogarithmicTransformation() {
      var logarithmicTransformation = (LogarithmicTransformation)transformation;
      var bValue = logarithmicTransformation.Base;

      return GenTreePow_b_x(bValue);
    }


    private ISymbolicExpressionTree GenerateModelForPowerTransformation() {
      var powerTransformation = (PowerTransformation)transformation;
      var expValue = powerTransformation.Exponent;

      // x ^ exp
      var powerNode = new Power().CreateTreeNode();
      var xNode = new Variable(column, "x").CreateTreeNode();
      var expNode = new ConstantTreeNode(new Constant() { Description = "exp" }) { Value = expValue };
      powerNode.AddSubtree(xNode);
      powerNode.AddSubtree(expNode);

      tree.Root.AddSubtree(powerNode);
      return tree;
    }

    private ISymbolicExpressionTree GenerateInverseModelForPowerTransformation() {
      var powerTransformation = (PowerTransformation)transformation;
      var expValue = powerTransformation.Exponent;

      // rt(x, b)
      var rootNode = new Root().CreateTreeNode();
      var xNode = new Variable(column, "x").CreateTreeNode();
      var bNode = new ConstantTreeNode(new Constant() { Description = "b" }) { Value = expValue };
      rootNode.AddSubtree(xNode);
      rootNode.AddSubtree(bNode);

      tree.Root.AddSubtree(rootNode);
      return tree;
    }


    private ISymbolicExpressionTree GenerateModelForReciprocalTransformation() {
      return GenTreeDiv_1_x();
    }

    private ISymbolicExpressionTree GenerateInverseModelForReciprocalTransformation() {
      return GenTreeDiv_1_x();
    }

    private ISymbolicExpressionTree GenerateModelForShiftStandardDistributionTransformation() {
      var shiftStandardDistributionTransformation = (ShiftStandardDistributionTransformation)transformation;
      var m_orgValue = shiftStandardDistributionTransformation.OriginalMean;
      var s_orgValue = shiftStandardDistributionTransformation.OriginalStandardDeviation;
      var m_tarValue = shiftStandardDistributionTransformation.Mean;
      var s_tarValue = shiftStandardDistributionTransformation.StandardDeviation;

      // x - m_org
      var substractionNode = new Subtraction().CreateTreeNode();
      var xNode = new Variable(column, "x").CreateTreeNode();
      var m_orgNode = new ConstantTreeNode(new Constant() { Description = "m_org" }) { Value = m_orgValue };
      substractionNode.AddSubtree(xNode);
      substractionNode.AddSubtree(m_orgNode);

      // (x - m_org) / s_org
      var divisionNode = new Division().CreateTreeNode();
      var s_orgNode = new ConstantTreeNode(new Constant() { Description = "s_org" }) { Value = s_orgValue };
      divisionNode.AddSubtree(substractionNode);
      divisionNode.AddSubtree(s_orgNode);

      // ((x - m_org) / s_org ) * s_tar
      var multiplicationNode = new Multiplication().CreateTreeNode();
      var s_tarNode = new ConstantTreeNode(new Constant() { Description = "s_tar" }) { Value = s_tarValue };
      multiplicationNode.AddSubtree(divisionNode);
      multiplicationNode.AddSubtree(s_tarNode);

      // ((x - m_org) / s_org ) * s_tar + m_tar
      var additionNode = new Addition().CreateTreeNode();
      var m_tarNode = new ConstantTreeNode(new Constant() { Description = "m_tar" }) { Value = m_tarValue };
      additionNode.AddSubtree(multiplicationNode);
      additionNode.AddSubtree(m_tarNode);

      tree.Root.AddSubtree(additionNode);
      return tree;
    }

    // helper's helper:

    private ISymbolicExpressionTree GenTreeLog_x_b(double b) {
      var tree = CreateNewTree();

      // log(x, b)
      var logNode = new Logarithm().CreateTreeNode();
      var bNode = new ConstantTreeNode(new Constant() { Description = "b" }) { Value = b };
      var xNode = new Variable(column, "x").CreateTreeNode();
      logNode.AddSubtree(xNode);
      logNode.AddSubtree(bNode);

      tree.Root.AddSubtree(logNode);
      return tree;
    }

    private ISymbolicExpressionTree GenTreePow_b_x(double b) {
      var tree = CreateNewTree();

      // b ^ x
      var powerNode = new Power().CreateTreeNode();
      var bNode = new ConstantTreeNode(new Constant() { Description = "b" }) { Value = b };
      var xNode = new Variable(column, "x").CreateTreeNode();
      powerNode.AddSubtree(bNode);
      powerNode.AddSubtree(xNode);

      tree.Root.AddSubtree(powerNode);
      return tree;
    }

    private ISymbolicExpressionTree GenTreeDiv_1_x() {
      var tree = CreateNewTree();

      // 1 / x
      var divNode = new Division().CreateTreeNode();
      var oneNode = new ConstantTreeNode(new Constant() { Description = "1" }) { Value = 1.0 };
      var xNode = new Variable(column, "x").CreateTreeNode();
      divNode.AddSubtree(oneNode);
      divNode.AddSubtree(xNode);

      tree.Root.AddSubtree(divNode);
      return tree;
    }


    private ISymbolicExpressionTree CreateNewTree() {
      return new SymbolicExpressionTree(new ProgramRootSymbol().CreateTreeNode());
    }

    private void InitComponents(ITransformation transformation) {
      this.transformation = transformation;
      column = transformation.Column;
      tree = CreateNewTree();
    }
  }
}
